/**
  *  \file iterator.h
  *  \brief Iterator Tricks
  *  \author Stefan Reuther
  */
#ifndef STREU_CPLUSLIB_ITERATOR_H_INCLUDED
#define STREU_CPLUSLIB_ITERATOR_H_INCLUDED

#include <iterator>

/** Dereference-a-pointer meta-program.
    dereference_type<T*>::value_type = T. */
    
template<typename T>
struct dereference_type { };

template<typename T>
struct dereference_type<T*> { typedef T value_type; };

/** Dereferencing iterator. If T is an iterator whose value_type is a pointer, i.e.
    "U*", dereferencing_iterator<T> iterates over the U's. */
template<typename T>
class dereferencing_iterator
    : public std::iterator<typename std::iterator_traits<T>::iterator_category,
                           typename dereference_type<typename std::iterator_traits<T>::value_type>::value_type,
                           typename std::iterator_traits<T>::difference_type,
                           typename std::iterator_traits<T>::pointer,
                           typename std::iterator_traits<T>::reference>
{
    typedef typename dereference_type<typename std::iterator_traits<T>::value_type>::value_type my_value_type;
    T iter;
 public:
    dereferencing_iterator()
        : iter()
        { }
    dereferencing_iterator(T iter)
        : iter(iter)
        { }

    bool operator==(const dereferencing_iterator& other) const
        { return iter == other.iter; }
    bool operator!=(const dereferencing_iterator& other) const
        { return iter != other.iter; }
    bool operator<(const dereferencing_iterator& other) const
        { return iter < other.iter; }
    bool operator<=(const dereferencing_iterator& other) const
        { return iter <= other.iter; }
    bool operator>(const dereferencing_iterator& other) const
        { return iter > other.iter; }
    bool operator>=(const dereferencing_iterator& other) const
        { return iter >= other.iter; }

    my_value_type& operator*() const
        { return **iter; }
    my_value_type* operator->() const
        { return *iter; }

    dereferencing_iterator& operator++()
        { ++iter; return *this; }
    dereferencing_iterator operator++(int)
        { dereferencing_iterator tmp(*this); ++iter; return tmp; }
    dereferencing_iterator& operator--()
        { --iter; return *this; }
    dereferencing_iterator operator--(int)
        { dereferencing_iterator tmp(*this); --iter; return tmp; }

    dereferencing_iterator& operator+=(typename std::iterator_traits<T>::difference_type d)
        { iter += d; return *this; }
    dereferencing_iterator operator+(typename std::iterator_traits<T>::difference_type d)
        { return dereferencing_iterator(iter + d); }
    dereferencing_iterator& operator-=(typename std::iterator_traits<T>::difference_type d)
        { iter -= d; return *this; }
    dereferencing_iterator operator-(typename std::iterator_traits<T>::difference_type d)
        { return dereferencing_iterator(iter + d); }
    typename std::iterator_traits<T>::difference_type operator-(const dereferencing_iterator& other)
        { return iter - other.iter; }
    my_value_type& operator[](typename std::iterator_traits<T>::difference_type d)
        { return *iter[d]; }
};

/** Member iterator. If T is an iterator whose value_type is a structure,
    this iterates over a particular member of all these structures. For
    example, <tt>member_iterator<std::map<int,int>::iterator,int,&std::pair<const int,int>::second></tt>
    iterates over all values in the map.
    \param T   Iterator type
    \param M   Member type
    \param Mem Member */
template<typename T, typename M, M (T::value_type::*Mem)>
class member_iterator
    : public std::iterator<typename std::iterator_traits<T>::iterator_category,
                           M,
                           typename std::iterator_traits<T>::difference_type,
                           typename std::iterator_traits<T>::pointer,
                           typename std::iterator_traits<T>::reference>
{
    typedef M my_value_type;
    T iter;
 public:
    member_iterator()
        : iter()
        { }
    member_iterator(T iter)
        : iter(iter)
        { }

    bool operator==(const member_iterator& other) const
        { return iter == other.iter; }
    bool operator!=(const member_iterator& other) const
        { return iter != other.iter; }
    bool operator<(const member_iterator& other) const
        { return iter < other.iter; }
    bool operator<=(const member_iterator& other) const
        { return iter <= other.iter; }
    bool operator>(const member_iterator& other) const
        { return iter > other.iter; }
    bool operator>=(const member_iterator& other) const
        { return iter >= other.iter; }

    my_value_type& operator*() const
        { return (*iter).*Mem; }
    my_value_type* operator->() const
        { return &(*iter).*Mem; }

    member_iterator& operator++()
        { ++iter; return *this; }
    member_iterator operator++(int)
        { member_iterator tmp(*this); ++iter; return tmp; }
    member_iterator& operator--()
        { --iter; return *this; }
    member_iterator operator--(int)
        { member_iterator tmp(*this); --iter; return tmp; }

    member_iterator& operator+=(typename std::iterator_traits<T>::difference_type d)
        { iter += d; return *this; }
    member_iterator operator+(typename std::iterator_traits<T>::difference_type d)
        { return member_iterator(iter + d); }
    member_iterator& operator-=(typename std::iterator_traits<T>::difference_type d)
        { iter -= d; return *this; }
    member_iterator operator-(typename std::iterator_traits<T>::difference_type d)
        { return member_iterator(iter + d); }
    typename std::iterator_traits<T>::difference_type operator-(const member_iterator& other)
        { return iter - other.iter; }
    my_value_type& operator[](typename std::iterator_traits<T>::difference_type d)
        { return (*iter[d]).*Mem; }
};

#endif
